# ------------------------------------------------------------------------------
# Name:          HumdrumFileStructure.py
# Purpose:       Responsible for rhythm analysis, parameter analysis (including
#                RDF signifier analysis), null token resolution, strand analysis,
#                and strophe analysis.
#
# Authors:       Greg Chapman <gregc@mac.com>
#                Humdrum code derived/translated from humlib (authored by
#                       Craig Stuart Sapp <craig@ccrma.stanford.edu>)
#
# Copyright:     (c) 2021-2022 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import sys
import typing as t
from operator import attrgetter
from fractions import Fraction
from pathlib import Path

from music21.common import opFrac

from converter21.humdrum import HumdrumInternalError
from converter21.humdrum import HumNum, HumNumIn
from converter21.humdrum import Convert
from converter21.humdrum import HumdrumToken
from converter21.humdrum import HumdrumLine
from converter21.humdrum import HumdrumFileBase
from converter21.humdrum import TokenPair

class HumdrumFileStructure(HumdrumFileBase):
    # HumdrumFileStructure has no private data to initialize, just a bunch more functions
    # So... no __init__
    # Except... pylint is unhappy about some data it thinks is uninitialized (because it's
    # initialized in HumdrumFileBase), so let's initialize it here, too.
    def __init__(self, fileName: t.Optional[t.Union[str, Path]] = None) -> None:
        super().__init__(fileName)
        self._ticksPerQuarterNote: int = -1
        self._barlines: t.List[HumdrumLine] = []
        self._strand1d: t.List[TokenPair] = []
        self._strand2d: t.List[t.List[TokenPair]] = []

    def readString(self, contents: str) -> bool:
        if not super().readString(contents):
            return self.isValid
        return self.analyzeStructure()

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeStructure -- Analyze global/local
    //    parameters and rhythmic structure.
    '''
    def analyzeStructure(self) -> bool:
        # unconditionally re-analyzes everything (except strands, for some reason)
        # called from various tools, not just from self.readString()
        success = self.analyzeStructureNoRhythm()
        if not success:
            return self.isValid

        success = self.analyzeRhythmStructure()
        if not success:
            return self.isValid

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeStructureNoRhythm -- Analyze global/local
    //    parameters but not rhythmic structure.
    '''
    def analyzeStructureNoRhythm(self) -> bool:
        # called from analyzeStructure unconditionally, and from analyzeRhythmStructure if necessary
        self._analyses.structureAnalyzed = True

        if not self.areStrandsAnalyzed:
            success = self.analyzeStrands()
            if not success:
                return self.isValid

        success = self.analyzeGlobalParameters()
        if not success:
            return self.isValid

        success = self.analyzeLocalParameters()
        if not success:
            return self.isValid

        success = self.analyzeTokenDurations()
        if not success:
            return self.isValid

        self.analyzeSignifiers()

        return self.isValid

    '''
    /////////////////////////////
    //
    // HumdrumFileStructure::analyzeRhythmStructure --
    '''
    def analyzeRhythmStructure(self) -> bool:
        # called from HumdrumLine, not just from self.analyzeStructure(),
        # thus the need to analyzeStructureNoRhythm, if necessary
        self._analyses.rhythmAnalyzed = True
        self.setLinesRhythmAnalyzed()

        if not self.isStructureAnalyzed:
            success = self.analyzeStructureNoRhythm()
            if not success:
                return self.isValid

        firstSpine: t.Optional[HumdrumToken] = self.spineStartList[0]
        if firstSpine is not None and firstSpine.isDataType('**recip'):
            self.assignRhythmFromRecip(firstSpine)
        else:
            success = self.analyzeRhythm()
            if not success:
                return self.isValid
            success = self.analyzeDurationsOfNonRhythmicSpines()
            if not success:
                return self.isValid
        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::getScoreDuration -- Return the total duration
    //    of the score in quarter note units.  Returns zero if no lines in the
    //    file, or -1 if there are lines, but no rhythmic analysis has been done.
    '''
    @property
    def scoreDuration(self) -> HumNum:
        if self.lineCount == 0:
            return 0
        return opFrac(
            # BUGFIX: Added last line duration
            self._lines[-1].durationFromStart + self._lines[-1].duration
        )

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::tpq -- "Ticks per Quarter-note".  Returns the minimal
    //    number of integral time units that divide a quarter note into equal
    //    subdivisions.  This value is needed to convert Humdrum data into
    //    MIDI file data, MuseData, and MusicXML data.  Also useful for timebase
    //    type of operations on the data and describing the durations in terms
    //    of integers rather than with fractions.  This function will also
    //    consider the implicit durations of non-rhythmic spine data.
    '''
    def tpq(self) -> int:
        if self._ticksPerQuarterNote > 0:
            return self._ticksPerQuarterNote

        durationSet: t.Set[Fraction] = self.getPositiveLineDurationFractions()
        denoms: t.List[int] = []
        for dur in durationSet:
            if dur.denominator > 1:
                denoms.append(dur.denominator)

        lcm: int = 1
        if len(denoms) > 0:
            lcm = Convert.getLcm(denoms)
        self._ticksPerQuarterNote = lcm
        return self._ticksPerQuarterNote

    def getPositiveLineDurationFractions(self) -> t.Set[Fraction]:
        output: t.Set[Fraction] = set()
        for line in self._lines:
            if line.duration > 0:
                output.add(Fraction(line.duration))
        return output

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::assignRhythmFromRecip --
    '''
    def assignRhythmFromRecip(self, spineStart: HumdrumToken) -> bool:
        currTok: t.Optional[HumdrumToken] = spineStart
        while currTok is not None:
            if not currTok.isData:
                currTok = currTok.nextToken0
                continue

            if currTok.isNull:
                # This should not occur in a well-formed **recip spine, but
                # treat as a zero duration.
                currTok = currTok.nextToken0
                continue

            currTok.ownerLine.duration = Convert.recipToDuration(currTok.text)
            currTok = currTok.nextToken0

        # now go back and set the absolute position from the start of the file.
        totalDurSoFar: HumNum = opFrac(0)
        for line in self._lines:
            line.durationFromStart = totalDurSoFar
            if line.duration is None or line.duration < 0:
                line.duration = opFrac(0)
            totalDurSoFar = opFrac(totalDurSoFar + line.duration)

        # Analyze durations to/from barlines:
        success = self.analyzeMeter()
        if not success:
            return False
        success = self.analyzeNonNullDataTokens()
        if not success:
            return False

        return True


    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::setLineRhythmAnalyzed --
    '''
    def setLinesRhythmAnalyzed(self) -> None:
        for line in self._lines:
            line.rhythmAnalyzed = True

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeRhythm -- Analyze the rhythmic structure
    //     of the data.  Returns false if there was a parse error.
    '''
    def analyzeRhythm(self) -> bool:
        # self.setLinesRhythmAnalyzed() # already happened in analyzeRhythmStructure
        if self.maxTrack == 0:
            return True


        trackStartTok: t.Optional[HumdrumToken] = self.trackStart(1)
        if trackStartTok is None:
            return False

        startLine: int = trackStartTok.lineIndex
        success: bool

        for i in range(1, self.maxTrack + 1):
            trackStartTok = self.trackStart(i)
            if trackStartTok is None:
                continue
            if not trackStartTok.hasRhythm:
                # Can't analyze rhythm of spines that do not have rhythm
                continue
            if trackStartTok.lineIndex == startLine:
                success = self.assignDurationsToTrack(trackStartTok, 0)
                if not success:
                    return False
            else:
                # Spine does not start at beginning of data, so
                # the starting position of the spine has to be
                # determined before continuing.  Search for a token
                # which is on a line with assigned duration, then work
                # outwards from that position.
                continue

        # Go back and analyze spines that do not start at the
        # beginning of the data stream.
        for i in range(1, self.maxTrack + 1):
            trackStartTok = self.trackStart(i)
            if trackStartTok is None:
                continue
            if not trackStartTok.hasRhythm:
                # Can't analyze rhythm of spines that do not have rhythm
                continue
            if trackStartTok.lineIndex > startLine:
                success = self.analyzeRhythmOfFloatingSpine(trackStartTok)
                if not success:
                    return False

        success = self.analyzeNullLineRhythms()
        if not success:
            return False
        self.fillInMissingStartTimes()
        self.assignLineDurations()
        success = self.analyzeMeter()
        if not success:
            return False
        success = self.analyzeNonNullDataTokens()
        return success

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeMeter -- Store the times from the last barline
    //     to the current line, as well as the time to the next barline.
    //     the sum of these two will be the duration of the barline, except
    //     for barlines, where the getDurationToBarline() will store the
    //     duration of the measure staring at that barline.  To get the
    //     beat, you will have to figure out the current time signature.
    '''
    def analyzeMeter(self) -> bool:
        self._barlines = []

        durationSum: HumNum = opFrac(0)
        foundFirstBarline: bool = False

        for line in self._lines:
            line.durationFromBarline = durationSum
            durationSum = opFrac(durationSum + line.duration)
            if line.isBarline:
                foundFirstBarline = True
                self._barlines.append(line)
                durationSum = opFrac(0)
            elif line.isData and not foundFirstBarline:
                # pickup measure, so set the first barline to the start of the file
                self._barlines.append(self._lines[0])
                foundFirstBarline = True

        durationSum = opFrac(0)
        for line in reversed(self._lines):
            durationSum = opFrac(durationSum + line.duration)
            line.durationToBarline = durationSum
            if line.isBarline:
                durationSum = opFrac(0)

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeTokenDurations -- Calculate the duration of
    //   all tokens in spines which posses duration in a file.
    '''
    def analyzeTokenDurations(self) -> bool:
        for line in self._lines:
            success = line.analyzeTokenDurations()
            if not success:
                return self.isValid
        return self.isValid

    '''
    ///////////////////////////////
    //
    // HumdrumFileStructure::analyzeGlobalParameters -- only allowing layout
    //    parameters at the moment.  Global parameters affect the next
    //    line which is either a barline, dataline or an interpretation
    //    other than a spine manipulator.  Null lines are also not
    //    considered.
    '''
    def analyzeGlobalParameters(self) -> bool:
        globalParamLines: t.List[HumdrumLine] = []

        for line in self._lines:
            if line.isGlobalComment and line.text.startswith('!!LO:'):
                line.storeGlobalLinkedParameters()  # stores those params in zeroth token of line
                globalParamLines.append(line)
                continue

            if not line.hasSpines:
                continue
            if line.isAllNull:
                continue
            if line.isLocalComment:
                continue
            if not globalParamLines:
                continue

            # Filter manipulators or not?  At the moment allow
            # global parameters to pass through manipulators.
            # if (m_lines[i]->isManipulator()) {
            # 	continue;
            # }

            for token in line.tokens():
                for gLine in globalParamLines:
                    # this token is affected by this global param
                    token.addLinkedParameterSet(gLine[0])

            globalParamLines = []

        return self.isValid

    '''
    ///////////////////////////////
    //
    // HumdrumFileStructure::analyzeLocalParameters -- Parses any
    //    local comments before a non-null token.
    '''
    def analyzeLocalParameters(self) -> bool:
        for i in range(0, self.strandCount()):
            self.processLocalParametersForStrand(i)

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeDurationsOfNonRhythmicSpines -- Calculate the
    //    duration of non-null data token in non-rhythmic spines.
    '''
    def analyzeDurationsOfNonRhythmicSpines(self) -> bool:
        for track in range(1, self.maxTrack + 1):  # tracks are 1-based
            for endIdx in range(0, self.trackEndCount(track)):
                trackEnd: t.Optional[HumdrumToken] = self.trackEnd(track, endIdx)
                if trackEnd is None or trackEnd.hasRhythm:
                    continue
                success = self.assignDurationsToNonRhythmicTrack(trackEnd, trackEnd)
                if not success:
                    return self.isValid

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::assignDurationsToTrack -- Assign duration from starts
    //    for each rhythmic spine in the file.  Analysis is done recursively, one
    //    sub-spine at a time.  Duplicate analyses are prevented by the state
    //    variable in the HumdrumToken (currently called rhycheck because it is only
    //    used in this function).  After the durationFromStarts have been assigned
    //    for the rhythmic analysis of non-data tokens and non-rhythmic spines is
    //    done elsewhere.
    '''
    def assignDurationsToTrack(self, startToken: HumdrumToken, startDur: HumNumIn) -> bool:
        sDur: HumNum = opFrac(startDur)
        if not startToken.hasRhythm:
            return self.isValid

        success: bool = self.prepareDurations(startToken, startToken.rhythmAnalysisState, sDur)
        if not success:
            return self.isValid
        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::prepareDurations -- Helper function for
    //     HumdrumFileStructure::assignDurationsToTrack() which does all of the
    //     work for assigning durationFromStart values.
    '''
    def prepareDurations(self, token: HumdrumToken, state: int, startDur: HumNumIn) -> bool:
        if state != token.rhythmAnalysisState:
            return self.isValid

        token.incrementRhythmAnalysisState()

        durSum: HumNum = opFrac(startDur)

        success = self.setLineDurationFromStart(token, durSum)
        if not success:
            return self.isValid

        if token.duration > 0:
            durSum = opFrac(durSum + token.duration)

        reservoir: t.List[HumdrumToken] = []
        startDurs: t.List[HumNum] = []

        # Assign line durationFromStarts for primary track first
        tcount: int = token.nextTokenCount
        while tcount > 0:
            for i, tok in enumerate(token.nextTokens):
                if i == 0:
                    # we'll deal with token 0 ourselves below
                    continue
                reservoir.append(tok)
                startDurs.append(durSum)

            if t.TYPE_CHECKING:
                # we know here that token.nextTokenCount > 0, so
                # token.nextToken0 is not None
                assert isinstance(token.nextToken0, HumdrumToken)

            token = token.nextToken0
            if state != token.rhythmAnalysisState:
                break

            token.incrementRhythmAnalysisState()
            success = self.setLineDurationFromStart(token, durSum)
            if not success:
                return self.isValid

            if token.duration > 0:
                durSum = opFrac(durSum + token.duration)

            tcount = token.nextTokenCount

        if tcount == 0 and token.isTerminateInterpretation:
            success = self.setLineDurationFromStart(token, durSum)
            if not success:
                return self.isValid

        # Process secondary tracks next:
        newState: int = state
        for i in reversed(range(0, len(reservoir))):
            self.prepareDurations(reservoir[i], newState, startDurs[i])

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::setLineDurationFromStart -- Set the duration of
    //      a line based on the analysis of tokens in the spine.
    '''
    def setLineDurationFromStart(self, token: HumdrumToken, durSum: HumNumIn) -> bool:
        dSum: HumNum = opFrac(durSum)
        if not token.isTerminateInterpretation and token.duration < 0:
            # undefined rhythm, so don't assign line duration information
            return self.isValid

        line: HumdrumLine = token.ownerLine
        if line.durationFromStart == -1:
            line.durationFromStart = dSum
        elif line.durationFromStart != dSum:
            if not token.isTerminateInterpretation:
                return self.setParseError(
                    f'''Error: Inconsistent rhythm analysis occurring near line {token.lineNumber}
Expected durationFromStart to be: {dSum} but found it to be {line.durationFromStart}
Line: {line.text}'''
                )
            line.durationFromStart = max(line.durationFromStart, dSum)

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeRhythmOfFloatingSpine --  This analysis
    //    function is used to analyze the rhythm of spines which do not start at
    //    the beginning of the data.  The function searches for the first line
    //    which has an assigned durationFromStart value, and then uses that
    //    as the basis for assigning the initial durationFromStart position
    //    for the spine.
    '''
    def analyzeRhythmOfFloatingSpine(self, spineStart: HumdrumToken) -> bool:
        durSum: HumNum = opFrac(0)
        foundDur: HumNum = opFrac(0)

        # Find a known durationFromStart for a line in the Humdrum file, then
        # use that to calculate the starting duration of the floating spine.
        if spineStart.durationFromStart >= 0:
            foundDur = spineStart.durationFromStart
        else:
            token: t.Optional[HumdrumToken] = spineStart
            while token is not None:
                if token.durationFromStart >= 0:
                    foundDur = token.durationFromStart
                    break
                if token.duration > 0:
                    durSum = opFrac(durSum + token.duration)
                token = token.nextToken0

        if foundDur == 0:
            return self.setParseError('Error: cannot link floating spine to score.')

        success: bool = self.assignDurationsToTrack(spineStart, foundDur - durSum)
        if not success:
            return self.isValid

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeNullLineRhythms -- When a series of null-token
    //    data line occur between two data lines possessing a start duration,
    //    then split the duration between those two lines amongst the null-token
    //    lines.  For example if a data line starts at time 15, and there is one
    //    null-token line before another data line at time 16, then the null-token
    //    line will be assigned to the position 15.5 in the score.
    '''
    def analyzeNullLineRhythms(self) -> bool:
        nullLines: t.List[HumdrumLine] = []
        previousLine: t.Optional[HumdrumLine] = None
        nextLine: t.Optional[HumdrumLine] = None

        for line in self._lines:
            if t.TYPE_CHECKING:
                # we know that every element of self._lines is not None
                assert isinstance(line, HumdrumLine)

            if not line.hasSpines:
                continue

            if line.isBarline:
                # We start from scratch in each measure.  This is because, if there is a null data
                # line as the first line in a measure, we don't want it to start halfway from last
                # real note in previous measure to first real note in this measure.  Any such
                # unprocessed null lines will end up inheriting their start time from the first
                # non-null note in this measure, during fillInMissingStartTimes' first loop
                # (backwards) over the lines.
                previousLine = None
                nullLines = []

            if line.isAllRhythmicNull:
                if line.isData:
                    nullLines.append(line)
                continue

            if line.durationFromStart < 0:
                if line.isData:
                    return self.setParseError(
                        'Error: found an unexpectedly missing durationFromStart on '
                        + f'data line {line.durationFromStart}\n'
                        + f'Line: {line.text}'
                    )
                continue

            nextLine = line
            if previousLine is None:
                previousLine = nextLine
                nullLines = []
                continue

            if t.TYPE_CHECKING:
                # we know previousLine is not None if we get here
                assert isinstance(previousLine, HumdrumLine)

            startDur: HumNum = previousLine.durationFromStart
            endDur: HumNum = nextLine.durationFromStart
            gapDur: HumNum = opFrac(endDur - startDur)
            nullDur: HumNum = opFrac(gapDur / (len(nullLines) + 1))
            for j, nullLine in enumerate(nullLines):
                nullLine.durationFromStart = opFrac(startDur + opFrac(nullDur * (j + 1)))

            previousLine = nextLine
            nullLines = []

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::fillInNegativeStartTimes -- Negative line durations
    //    after the initial rhythmAnalysis mean that the lines are not data line.
    //    Duplicate the duration of the next non-negative duration for all negative
    //    durations.
        "negative" is now "None" -- gregc
    '''
    def fillInMissingStartTimes(self) -> None:
        lastDur: HumNum = opFrac(-1)

        for line in reversed(self._lines):
            if line.durationFromStart < 0 and lastDur >= 0:
                line.durationFromStart = lastDur
            if line.durationFromStart >= 0:
                lastDur = line.durationFromStart

        # fill in start times for ending comments
        for line in self._lines:
            if line.durationFromStart >= 0:
                lastDur = line.durationFromStart
            else:
                line.durationFromStart = lastDur

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::assignLineDurations --  Calculate the duration of lines
    //   based on the durationFromStart of the current line and the next line.
    '''
    def assignLineDurations(self) -> None:
        for i in range(0, len(self._lines)):
            if i == len(self._lines) - 1:
                self._lines[i].duration = opFrac(0)
            else:
                startDur: HumNum = self._lines[i].durationFromStart
                endDur: HumNum = self._lines[i + 1].durationFromStart
                self._lines[i].duration = opFrac(endDur - startDur)

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::assignDurationsToNonRhythmicTrack --  After the basic
    //   rhythmAnalysis has been done, go back and assign durations to non-rhythmic
    //   spine tokens based on the lineFromStart values of the lines that they
    //   occur on as well as the distance in the file to the next non-null token for
    //   that spine.
    '''
    def assignDurationsToNonRhythmicTrack(
            self,
            endToken: HumdrumToken,
            current: HumdrumToken
    ) -> bool:
        spineInfo: str = endToken.spineInfo
        token: t.Optional[HumdrumToken] = endToken

        while token is not None:
            if token.spineInfo != spineInfo:
                if 'b' in token.spineInfo:
                    break
                if 'b' in spineInfo:
                    break

            tcount: int = token.previousTokenCount
            if tcount == 0:
                break

            if tcount > 1:
                for i in range(1, tcount):
                    ptok: t.Optional[HumdrumToken] = token.previousToken(i)
                    if t.TYPE_CHECKING:
                        # we know that ptok is not None
                        assert isinstance(ptok, HumdrumToken)

                    success = self.assignDurationsToNonRhythmicTrack(ptok, current)
                    if not success:
                        return self.isValid

            if token.isNonNullData:
                token.duration = opFrac(current.durationFromStart - token.durationFromStart)
                current = token

            token = token.previousToken0

        return self.isValid

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::processLocalParametersForStrand --
    '''
    def processLocalParametersForStrand(self, index: int) -> None:
        sStart: t.Optional[HumdrumToken] = self.strandStart1d(index)
        sEnd: t.Optional[HumdrumToken] = self.strandEnd1d(index)
        tok: t.Optional[HumdrumToken] = sEnd  # start at the end and work backward
        dtok: t.Optional[HumdrumToken] = None

        while tok is not None:
            if tok.isData:
                dtok = tok
            elif tok.isBarline:
                # layout parameters allowed for barlines
                dtok = tok
            elif tok.isInterpretation and tok.text != '*' and not tok.isManipulator:
                # layout parameters allowed for non-null, non-manip interpretations
                dtok = tok
            elif tok.isLocalComment and tok.text.startswith('!LO:'):
                tok.storeParameterSet()
                if dtok is not None:
                    dtok.addLinkedParameterSet(tok)

            if tok == sStart:
                break

            tok = tok.previousToken0

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeStrands -- Analyze spine strands.
    '''
    def analyzeStrands(self) -> bool:
        self._analyses.strandsAnalyzed = True

        self._strand1d = []
        self._strand2d = []

        for i in range(0, self.spineCount):
            tok: t.Optional[HumdrumToken] = self.spineStartList[i]
            self._strand2d.append([])
            self.analyzeSpineStrands(self._strand2d[-1], tok)

        for i in range(0, len(self._strand2d)):
            self._strand2d[i].sort(key=attrgetter('firstLineIndex', 'firstFieldIndex'))
            for j in range(0, len(self._strand2d[i])):
                self._strand1d.append(self._strand2d[i][j])

        self.assignStrandsToTokens()
        self.resolveNullTokens()
        self.analyzeLocalParameters()
#        self.analyzeStrophes()

        return self.isValid

    '''
    ///////////////////////////////
    //
    // HumdrumFileStructure::resolveNullTokens --
    '''
    def resolveNullTokens(self) -> None:
        if self._analyses.nullsAnalyzed:
            return

        self._analyses.nullsAnalyzed = True
        if not self.areStrandsAnalyzed:
            self.analyzeStrands()

        data: t.Optional[HumdrumToken] = None
        strandPair: TokenPair
        for strandPair in self._strand1d:
            token: t.Optional[HumdrumToken] = strandPair.first
            strandEnd: t.Optional[HumdrumToken] = strandPair.last
            if t.TYPE_CHECKING:
                # after analyzeStrands, no pair in self._strand1d will have Nones in it.
                assert token is not None
                assert strandEnd is not None

            while token != strandEnd:
                if token is None:
                    # we never reached strandEnd, but we're done
                    raise HumdrumInternalError(f'never found strandEnd ({strandEnd})')

                if not token.isData:
                    token = token.nextToken0
                    continue

                if data is None:
                    data = token
                    token.nullResolution = data
                    token = token.nextToken0
                    continue

                if token.isNull:
                    token.nullResolution = data
                else:
                    data = token

                token = token.nextToken0

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::assignStrandsToTokens -- Store the 1D strand
    //    index number for each token in the file.  Global tokens will have
    //    strand index set to -1.
    '''
    def assignStrandsToTokens(self) -> None:
        strandPair: TokenPair
        for i, strandPair in enumerate(self._strand1d):
            tok: t.Optional[HumdrumToken] = strandPair.first
            while tok is not None:
                tok.strandIndex = i
                tok = tok.nextToken0

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeSpineStrands -- Fill in the list of
    //   strands in a single spine.
    '''
    def analyzeSpineStrands(
            self,
            ends: t.List[TokenPair],
            startToken: t.Optional[HumdrumToken]
    ) -> None:
        newStrand: TokenPair = TokenPair(startToken, None)
        ends.append(newStrand)

        tok: t.Optional[HumdrumToken] = startToken
        while tok is not None:
            if tok.isMergeInterpretation and tok.subTrack > 1:
                # check to the left: if the left primary/sub spine also has
                # a *v, then this is the end of this strand; otherwise, the
                # strand continues.
                if tok.previousFieldToken is not None:
                    if tok.previousFieldToken.isMergeInterpretation:
                        newStrand.last = tok
                        return

                tok = tok.nextToken0
                continue

            if tok.isTerminateInterpretation:
                newStrand.last = tok
                return

            if tok.nextTokenCount > 1:
                # should only be 2, but allow for generalizing in the future.
                for j in range(1, tok.nextTokenCount):
                    self.analyzeSpineStrands(ends, tok.nextToken(j))

            tok = tok.nextToken0

        print('Should not get here in analyzeSpineStrands()', file=sys.stderr)

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::getStrandCount --
    '''
    def strandCount(self, spineIndex: int = None) -> int:
        if not self.areStrandsAnalyzed:
            self.analyzeStrands()

        if spineIndex is None:
            # caller is asking for strand count of entire file
            return len(self._strand1d)

        # caller is asking for strand count of a particular spine
        if spineIndex < 0 or spineIndex >= len(self._strand2d):
            return 0

        return len(self._strand2d[spineIndex])

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::getStrandStart -- Return the first token
    //    in the a strand.
    '''
    def strandStart2d(self, spineIndex: int, strandIndex: int) -> t.Optional[HumdrumToken]:
        if not self.areStrandsAnalyzed:
            self.analyzeStrands()

        return self._strand2d[spineIndex][strandIndex].first

    def strandStart1d(self, strandIndex: int) -> t.Optional[HumdrumToken]:
        if not self.areStrandsAnalyzed:
            self.analyzeStrands()

        return self._strand1d[strandIndex].first

    def strandEnd2d(self, spineIndex: int, strandIndex: int) -> t.Optional[HumdrumToken]:
        if not self.areStrandsAnalyzed:
            self.analyzeStrands()

        return self._strand2d[spineIndex][strandIndex].last

    def strandEnd1d(self, strandIndex: int) -> t.Optional[HumdrumToken]:
        if not self.areStrandsAnalyzed:
            self.analyzeStrands()

        return self._strand1d[strandIndex].last

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::hasFilters -- Returns true if has any
    //    reference records starting with "!!!filter:" or "!!!!filter:".
    '''
    def hasFilters(self) -> bool:
        refs: t.List[HumdrumLine] = self.globalReferenceRecords()
        for ref in refs:
            if ref.globalReferenceKey == 'filter':
                return True
        return False

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::hasGlobalFilters -- Returns true if has any
    //    reference records starting with "!!!filter:".
    '''
    def hasGlobalFilters(self) -> bool:
        for line in self._lines:
            if not line.isComment:
                continue

            token0: t.Optional[HumdrumToken] = line[0]
            if token0 is not None and token0.text.startswith('!!!filter'):
                return True

        return False

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::hasUniversalFilters -- Returns true if has any
    //    reference records starting with "!!!!filter:".
    '''
    def hasUniversalFilters(self) -> bool:
        refs: t.List[HumdrumLine] = self.universalReferenceRecords()
        for ref in refs:
            if ref.universalReferenceKey == 'filter':
                return True
        return False

    '''
    //////////////////////////////
    //
    // HumdrumFileStructure::analyzeSignifiers --
        Analyzes all the RDF lines
    '''
    def analyzeSignifiers(self) -> None:
        for line in self._lines:
            if line.isSignifier:
                self._signifiers.addSignifier(line.text)
        self._signifiers.generateKnownInfo()



# ------------------------------------------------------------------------------
# Name:          HumdrumTools.py
# Purpose:       HumdrumTools contains various tools that modify Humdrum data.
#
# Authors:       Greg Chapman <gregc@mac.com>
#                Humdrum code derived/translated from humlib (authored by
#                       Craig Stuart Sapp <craig@ccrma.stanford.edu>)
#
# Copyright:     (c) 2021-2022 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------

import sys
import re
import math
import typing as t
from fractions import Fraction

from music21.common import opFrac

from converter21.humdrum import HumdrumFile
from converter21.humdrum import HumdrumLine
from converter21.humdrum import HumdrumToken
from converter21.humdrum import HumNum
from converter21.humdrum import Convert

class ToolTremolo:
    # Tremolo expansion tool.
    def __init__(self, infile: HumdrumFile) -> None:
        self.infile: HumdrumFile = infile
        self.markupTokens: t.List[HumdrumToken] = []
        self.firstTremoloLinesInTrack: t.List[t.List[t.Optional[HumdrumLine]]] = []
        self.lastTremoloLinesInTrack: t.List[t.List[t.Optional[HumdrumLine]]] = []

    '''
    //////////////////////////////
    //
    // Tool_tremolo::processFile --
    '''
    def processFile(self) -> None:
        if self.infileNeedsToBeParsed():
            self.infile.analyzeBase()
            self.infile.analyzeStructure()
            if not self.infile.isValid:
                return

        for _ in range(0, self.infile.maxTrack + 1):
            self.firstTremoloLinesInTrack.append([])
            self.lastTremoloLinesInTrack.append([])

        for line in reversed(list(self.infile.lines())):
            if not line.isData:
                continue
            if line.duration == 0:
                # don't deal with grace notes
                continue

            for token in line.tokens():
                if not token.isKern:
                    continue
                if token.isNull:
                    continue

                m = re.search(r'@(\d+)@', token.text)
                if not m:
                    continue

                self.markupTokens.insert(0, token)  # markupTokens is in forward order

                value: int = int(m.group(1))
                duration: HumNum = Convert.recipToDuration(token.text)
                four: HumNum = opFrac(4)
                count: HumNum = opFrac((duration * value) / four)
                increment: HumNum = opFrac(four / value)

                if '@@' in token.text:
                    count = opFrac(count * 2)

                countFraction: Fraction = Fraction(count)
                if countFraction.denominator != 1:
                    print(f'Error: tremolo time value cannot be used: {value}', file=sys.stderr)
                    continue

                kcount: int = countFraction.numerator
                startTime: HumNum = token.durationFromStart
                for k in range(1, kcount):
                    timestamp: HumNum = opFrac(startTime + (increment * opFrac(k)))
                    self.infile.insertNullDataLine(timestamp)

        self.expandTremolos()
        self.addTremoloInterpretations()

        # is this necessary?  We do createLineFromTokens() whenever we modify a token already.
#         if self.modified:
#             self.infile.createLinesFromTokens()

    def infileNeedsToBeParsed(self) -> bool:
        if not self.infile.isStructureAnalyzed:
            return True
        if not self.infile.isRhythmAnalyzed:
            return True

        return False


    def notesFoundInTrackBetweenLineIndices(self, track: int, startIdx: int, endIdx: int) -> bool:
        # Check every line between startIdx and endIdx, NOT inclusive at either end
        for i in range(startIdx + 1, endIdx):
            line: t.Optional[HumdrumLine] = self.infile[i]
            if line is None:
                continue
            for token in line.tokens():
                if token.track == track:
                    if token.isNote:
                        return True
        return False

    '''
    //////////////////////////////
    //
    // Tool_tremolo::addTremoloInterpretations --
    '''
    def addTremoloInterpretations(self) -> None:
        # These starts/stops are exactly what music21 has, which is per beam-group.
        # We would like to avoid a stop immediately followed by a start (as redundant).
        # Or even a stop and then a start with no actual notes between them.
        # So we first "optimize" this.
        for track, (firstLines, lastLines) in enumerate(zip(
                self.firstTremoloLinesInTrack,
                self.lastTremoloLinesInTrack)):
            lastLineRemovalIndices: t.List[int] = []
            firstLineRemovalIndices: t.List[int] = []
            currLastLineIndex: t.Optional[int] = None
            prevLastLineIndex: t.Optional[int] = None
            currFirstLineIndex: t.Optional[int] = None
            for idx, (firstLine, lastLine) in enumerate(zip(firstLines, lastLines)):
                if t.TYPE_CHECKING:
                    # self.{first,last}TremoloLinesInTrack should be filled in by now
                    assert firstLine is not None
                    assert lastLine is not None

                prevLastLineIndex = currLastLineIndex
                currFirstLineIndex = firstLine.lineIndex
                currLastLineIndex = lastLine.lineIndex

                if prevLastLineIndex is not None:
                    if not self.notesFoundInTrackBetweenLineIndices(
                            track, prevLastLineIndex, currFirstLineIndex):
                        lastLineRemovalIndices.append(idx - 1)
                        firstLineRemovalIndices.append(idx)

            for lineIndexToRemove in reversed(lastLineRemovalIndices):
                lastLines.pop(lineIndexToRemove)
            for lineIndexToRemove in reversed(firstLineRemovalIndices):
                firstLines.pop(lineIndexToRemove)


        line: HumdrumLine

        # Insert starting *tremolo(s)
        for track, firstLines in enumerate(self.firstTremoloLinesInTrack):
            if not firstLines:
                # no first times in this track
                continue

            for firstLine in firstLines:
                if firstLine is None:
                    continue

                line = self.infile.insertNullInterpretationLineAt(
                    firstLine.lineIndex
                )
                if line is None:
                    continue

                for token in line.tokens():
                    if token.subTrack > 1:
                        # Currently *tremolo affects all subTracks, but this
                        # will probably change in the future.
                        continue
                    if token.track == track:
                        token.text = '*tremolo'
                        line.createLineFromTokens()

        # Insert ending *Xtremolo(s)
        for track, lastLines in enumerate(self.lastTremoloLinesInTrack):
            if not lastLines:
                continue

            for lastLine in lastLines:
                if lastLine is None:
                    continue

                line = self.infile.insertNullInterpretationLineAt(
                    lastLine.lineIndex + 1
                )
                if line is None:
                    continue

                for token in line.tokens():
                    if token.subTrack is not None and token.subTrack > 1:
                        # Currently *tremolo affects all subTracks, but this
                        # will probably change in the future.
                        continue
                    if token.track == track:
                        token.text = '*Xtremolo'
                        line.createLineFromTokens()

    '''
    //////////////////////////////
    //
    // Tool_tremolo::expandTremolos --
    '''
    def expandTremolos(self) -> None:
        for token in self.markupTokens:
            # if we have already replaced the '@' or '@@' token with an expansion note,
            # we just skip it here.
            if '@' not in token.text:
                continue

            if '@@' in token.text:
                self.expandFingerTremolo(token)
            else:
                self.expandTremolo(token)

    '''
    //////////////////////////////
    //
    // Tool_tremolo::expandTremolo --
    '''
    def expandTremolo(self, token: HumdrumToken) -> None:
        addBeam: bool = False
        tnotes: int = -1

        m = re.search(r'@(\d+)@', token.text)
        if not m:
            return

        value: int = int(m.group(1))
        valueHumNum: HumNum = opFrac(value)
        valueFraction: Fraction = Fraction(value)

        duration: HumNum = Convert.recipToDuration(token.text)

        four: HumNum = opFrac(4)
        count: HumNum = opFrac(duration * valueHumNum / four)
        countFraction: Fraction = Fraction(count)
        if countFraction.denominator != 1:
            print(f'Error: non-integer number of tremolo notes: {token}', file=sys.stderr)
            return
        if value < 8:
            print(f'Error: tremolo notes can only be eighth-notes or shorter: {token}',
                    file=sys.stderr)
            return
        if duration > 0.5:
            # needs to be less than one for tuplet quarter note tremolos
            addBeam = True

        # There are cases where duration < 1 need added beams
        # when the note is not already in a beam.  Such as
        # a plain 8th note with a slash.  This needs to be
        # converted into two 16th notes with a beam so that
        # *tremolo can reduce it back into a tremolo, since
        # it will only reduce beam groups.

        repeat: HumNum = opFrac((duration * valueHumNum) / four)
        increment: HumNum = opFrac(four / valueHumNum)
        repeatFraction: Fraction = Fraction(repeat)
        if repeatFraction.denominator != 1:
            print(f'Error: tremolo repetition count must be an integer: {token}',
                    file=sys.stderr)
            return
        tnotes = repeatFraction.numerator

        self.storeFirstTremoloNoteInfo(token)

        beams: int = int(math.log(float(value), 2)) - 2
        markup: str = f'@{valueFraction.numerator}@'
        base: str = re.sub(markup, '', token.text)

        # complicated beamings are not allowed yet (no internal L/J markers in tremolo beam)
        hasBeamStart: bool = 'L' in base
        hasBeamStop: bool = 'J' in base

        if addBeam:
            hasBeamStart = True
            hasBeamStop = True

        # Currently not allowed to add tremolo to beamed notes, so remove all beaming:
        base = re.sub(r'[LJKk]+', '', base)
        startBeam: str = 'L' * beams
        endBeam: str = 'J' * beams

        # Set the rhythm of the tremolo notes.
        # Augmentation dot is expected adjacent to regular rhythm value.
        # Maybe allow anywhere?
        base = re.sub(r'\d+%?\d*\.*', str(valueFraction.numerator), base)
        initial: str = base
        if hasBeamStart:
            initial += startBeam
        terminal: str = base
        if hasBeamStop:
            terminal += endBeam

        # remove tie continue from start of tremolo (leave tie start/end in place)
        initial = re.sub(r'_+[<>]?', '', initial)
        # remove tie information from middle of tremolo
        base = re.sub(r'[\[_\]]+[<>]?', '', base)
        # remove tie information from end of tremolo
        terminal = re.sub(r'[\[_\]]+[<>]?', '', terminal)

        # remove slur start from end of tremolo:
        terminal = re.sub(r'[(]+[<>]', '', terminal)

        token.text = initial
        token.ownerLine.createLineFromTokens()

        # Now fill in the rest of the tremolos.
        startTime: HumNum = token.durationFromStart
        timestamp: HumNum = opFrac(startTime + increment)
        currTok: t.Optional[HumdrumToken] = token.nextToken(0)
        counter: int = 1

        while currTok is not None:
            if not currTok.isData:
                currTok = currTok.nextToken(0)
                continue

            duration = currTok.ownerLine.duration
            if duration == 0:
                # grace note line, so skip
                currTok = currTok.nextToken(0)
                continue

            cstamp: HumNum = currTok.durationFromStart
            if cstamp < timestamp:
                currTok = currTok.nextToken(0)
                continue

            if cstamp > timestamp:
                print('\tWarning: terminating tremolo insertion early', file=sys.stderr)
                print(f'\tCSTAMP : {cstamp} TSTAMP : {timestamp}', file=sys.stderr)
                break

            counter += 1
            if counter == tnotes:
                currTok.text = terminal
                self.storeLastTremoloNoteInfo(currTok)
            else:
                currTok.text = base

            currTok.ownerLine.createLineFromTokens()
            if counter >= tnotes:
                # done with inserting of tremolo notes.
                break

            timestamp = opFrac(timestamp + increment)
            currTok = currTok.nextToken(0)

    '''
    //////////////////////////////
    //
    // Tool_tremolo::getNextNote --
    '''
    def getNextNote(self, token: HumdrumToken) -> t.Optional[HumdrumToken]:
        output: t.Optional[HumdrumToken] = None
        current: t.Optional[HumdrumToken] = token.nextToken0

        while current is not None:
            if not current.isData:
                current = current.nextToken0
                continue
            if current.duration == 0:
                # ignore grace notes
                current = current.nextToken0
                continue
            if current.isNull or current.isRest:
                current = current.nextToken0
                continue

            output = current
            break

        return output

    '''
    //////////////////////////////
    //
    // Tool_tremolo::expandFingerTremolos --
    '''
    def expandFingerTremolo(self, token1: HumdrumToken) -> None:
        token2: t.Optional[HumdrumToken] = self.getNextNote(token1)
        if token2 is None:
            return

        m = re.search(r'@@(\d+)@@', token1.text)
        if m is None:
            return

        value: int = int(m.group(1))
        valueHumNum: HumNum = opFrac(value)
        if not Convert.isPowerOfTwo(valueHumNum):
            print(f'Error: not a power of two: {token1}', file=sys.stderr)
            return
        if value < 8:
            print(f'Error: tremolo can only be eighth-notes or shorter: {token1}', file=sys.stderr)
            return

        duration: HumNum = Convert.recipToDuration(token1.text)
        four: HumNum = opFrac(4)
        count: HumNum = opFrac((duration * valueHumNum) / four)
        countFraction: Fraction = Fraction(count)

        if countFraction.denominator != 1:
            print(f'Error: tremolo repetition count must be an integer: {token1}', file=sys.stderr)
            return
        increment: HumNum = opFrac(four / valueHumNum)

        tnotes: int = countFraction.numerator * 2

        self.storeFirstTremoloNoteInfo(token1)

        beams: int = int(math.log(float(value), 2)) - 2
        markup: str = f'@@{value}@@'
        base1: str = token1.text
        base1 = re.sub(markup, '', base1)
        # Currently not allowed to add tremolo to beamed notes, so remove all beaming:
        base1 = re.sub(r'[LJKk]+', '', base1)
        startBeam: str = 'L' * beams
        endBeam: str = 'J' * beams

        # Set the rhythm of the tremolo notes.
        # Augmentation dot is expected adjacent to regular rhythm value.
        # Maybe allow anywhere?
        base1 = re.sub(r'\d+%?\d*\.*', str(value), base1)
        initial: str = base1 + startBeam
        # remove slur end from start of tremolo
        initial = re.sub(r'[)]+[<>]?', '', initial)
        # remove tie continue from start of tremolo (leave tie start/end in place)
        initial = re.sub(r'_+[<>]?', '', initial)

        # remove slur information from middle of tremolo
        base1 = re.sub(r'[()]+[<>]?', '', base1)
        # remove tie information from middle of tremolo
        base1 = re.sub(r'[\[_\]]+[<>]?', '', base1)

        token1.text = initial
        token1.ownerLine.createLineFromTokens()

        base2: str = token2.text
        base2 = re.sub(markup, '', base2)
        base2 = re.sub(r'[LJKk]+', '', base2)
        base2 = re.sub(r'\d+%?\d*\.*', str(value), base2)

        terminal: str = base2 + endBeam
        # remove slur start information from end of tremolo:
        terminal = re.sub(r'[(]+[<>]?', '', terminal)
        # remove tie information from end of tremolo
        terminal = re.sub(r'[\[_\]]+[<>]?', '', terminal)

        state: bool = False

        # Now fill in the rest of the tremolos.
        startTime: HumNum = token1.durationFromStart
        timestamp: HumNum = opFrac(startTime + increment)
        currTok: t.Optional[HumdrumToken] = token1.nextToken(0)
        counter: int = 1
        while currTok is not None:
            if not currTok.isData:
                currTok = currTok.nextToken(0)
                continue

            # We also skip zero-duration lines (grace notes)
            if currTok.ownerLine.duration == 0:
                currTok = currTok.nextToken(0)
                continue

            cstamp: HumNum = currTok.durationFromStart
            if cstamp < timestamp:
                currTok = currTok.nextToken(0)
                continue

            if cstamp > timestamp:
                print('\tWarning: terminating tremolo insertion early', file=sys.stderr)
                print(f'\tCSTAMP : {cstamp} TSTAMP : {timestamp}', file=sys.stderr)
                break

            counter += 1
            if counter == tnotes:
                currTok.text = terminal
                self.storeLastTremoloNoteInfo(currTok)
            else:
                if state:
                    currTok.text = base1
                else:
                    currTok.text = base2
                state = not state

            currTok.ownerLine.createLineFromTokens()

            if counter >= tnotes:
                # done with inserting of tremolo notes
                break

            timestamp = opFrac(timestamp + increment)
            currTok = currTok.nextToken(0)

    '''
    //////////////////////////////
    //
    // Tool_tremolo::storeFirstTremoloNote --
    '''
    def storeFirstTremoloNoteInfo(self, token: HumdrumToken) -> None:
        if token is None:
            return

        track: t.Optional[int] = token.track
        if track is None:
            print(f'Track is not set for token: {token}', file=sys.stderr)
            return

        self.firstTremoloLinesInTrack[track].append(token.ownerLine)

    '''
    //////////////////////////////
    //
    // Tool_tremolo::storeLastTremoloNote --
    '''
    def storeLastTremoloNoteInfo(self, token: HumdrumToken) -> None:
        if token is None:
            return

        track: t.Optional[int] = token.track
        if track is None:
            print(f'Track is not set for token: {token}', file=sys.stderr)
            return

        self.lastTremoloLinesInTrack[track].append(token.ownerLine)



# ------------------------------------------------------------------------------
# Name:          HumdrumFileContent.py
# Purpose:       Used to add content analysis to HumdrumFileStructure class,
#                and do other higher-level processing of Humdrum data.
#
# Authors:       Greg Chapman <gregc@mac.com>
#                Humdrum code derived/translated from humlib (authored by
#                       Craig Stuart Sapp <craig@ccrma.stanford.edu>)
#
# Copyright:     (c) 2021-2022 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import re
import typing as t
from fractions import Fraction
from pathlib import Path

from music21.common import opFrac

from converter21.humdrum import HumdrumInternalError
from converter21.humdrum import HumNum
from converter21.humdrum import Convert
from converter21.humdrum import HumdrumToken
from converter21.humdrum import HumdrumFileStructure
from converter21.humdrum import TokenPair

# spineColors array will contains slots for this many subtracks
MAXCOLORSUBTRACK: int = 30

class HumdrumFileContent(HumdrumFileStructure):
    # HumdrumFileContent has no private data to initialize, just a bunch more functions
    # So... no __init__.  Well, OK, to keep pylint happy about attribute-defined-outside-init,
    # a quick __init__ that sets a few attributes that are already set in HumdrumFileBase.py
    def __init__(self, fileName: t.Optional[t.Union[str, Path]] = None) -> None:
        super().__init__(fileName)  # initialize the HumdrumFileBase fields
        self._hasInformalBreaks: bool = False
        self._hasFormalBreaks: bool = False

        # spinecolor = self._spineColor[track][subtrack]
        self._spineColor: t.List[t.List[str]] = []

    def readString(self, contents: str) -> bool:
        if not super().readString(contents):
            return self.isValid
        return self.isValid

    def analyzeNotation(self) -> None:
        # This code originally was in verovio/iohumdrum.cpp (by Craig Sapp)
        # Here I am consolidating this into one HumdrumFileContent method which
        # can be called from any client that wants notation support (e.g.
        # HumdrumFile.createM21Stream)

        # Might be worth adding to HumdrumFileContent at some point
        # m_multirest = analyzeMultiRest(infile);

        self.analyzeBreaks()
        self.analyzeSlurs()
        self.analyzePhrasings()
        self.analyzeKernTies()
        # self.analyzeKernStemLengths()  # Don't know why this is commented out in iohumdrum.cpp
        self.analyzeRestPositions()
        self.analyzeKernAccidentals()
        self.analyzeTextRepetition()

#         if (m_signifiers.terminallong) {
#             hideTerminalBarlines(infile);
#         }
#         checkForColorSpine(infile); # interesting to move here
        self.analyzeRScale()

#         # iohumdrum calls this but it belongs in engraving, not in import
#         self.analyzeCrossStaffStemDirections();

#         # iohumdrum calls this to deal with measure vs barline in MEI.
#         self.analyzeBarlines();

        self.analyzeClefNulls()

#         if (infile.hasDifferentBarlines()) {
#             adjustMeasureTimings(infile); # Interesting to move here, but confuses me (why?)
#         }

        self.initializeSpineColor()
#         initializeIgnoreVector(infile); # interesting to move here

#         extractNullInformation(m_nulls, infile); # might be interesting?
#         Perhaps very MEI-specific.

    '''
    //////////////////////////////
    //
    // HumdrumInput::analyzeClefNulls -- Mark all null interpretations
    //    that are in the same track as a clef interpretation.
    '''
    def analyzeClefNulls(self) -> None:
        for line in self.lines():
            if not line.isInterpretation:
                continue

            for token in line.tokens():
                if not token.isKern:
                    continue
                if not token.isClef:
                    continue
                self.markAdjacentNullsWithClef(token)

    '''
    //////////////////////////////
    //
    // HumdrumInput::markAdjacentNullsWithClef -- Input is a clef token,
    //     and all null interpretations in the same track will be marked
    //     as being the same clef, since verovio/MEI requires clef changes
    //     to be present in all layers.
        ... and music21 likes this (in all voices) too. --gregc
    '''
    @staticmethod
    def markAdjacentNullsWithClef(clef: HumdrumToken) -> None:
        ctrack: t.Optional[int] = clef.track
        if ctrack is None:
            return

        track: t.Optional[int] = 0

        current: t.Optional[HumdrumToken] = clef.nextFieldToken
        while current is not None:
            track = current.track
            if track != ctrack:
                break
            if current.text == '*':
                current.setValue('auto', 'clef', clef.text)
            current = current.nextFieldToken

        current = clef.previousFieldToken
        while current is not None:
            track = current.track
            if track != ctrack:
                break
            if current.text == '*':
                current.setValue('auto', 'clef', clef.text)
            current = current.previousFieldToken

    '''
    //////////////////////////////
    //
    // HumdrumInput::analyzeBreaks -- Returns true if there are page or
    //   system breaks in the data.

        Actually sets self._hasFormalBreaks and self._hasInformalBreaks
    '''
    def analyzeBreaks(self) -> None:
        # check for informal breaking markers such as:
        # !!pagebreak:original
        # !!linebreak:original
        # we don't pay attention to informal breaks (these are the breaks that happened
        # to be in the original score, not breaks that were specified as part of that score)
        #   for line in self._lines:
        #       if not line.isGlobalComment:
        #           continue
        #
        #       if line[0].text.startswith('!!pagebreak:'):
        #           self._hasInformalBreaks = True
        #           break
        #
        #       if line[0].text.startswith('!!linebreak:'):
        #           self._hasInformalBreaks = True
        #           break

        # check for formal breaking markers such as:
        # !!LO:PB:g=z
        # !!LO:LB:g=z
        for line in self._lines:
            if not line.isComment:
                continue

            if line[0] is None:
                # shouldn't happen, but...
                continue

            if '!LO:LB' in line[0].text:
                self._hasFormalBreaks = True
                break

            if '!LO:PB' in line[0].text:
                self._hasFormalBreaks = True
                break

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeSlurs -- Link start and ends of
    //    slurs to each other.
    '''
    def analyzeSlurs(self) -> bool:
        if self._analyses.slursAnalyzed:
            return False

        self._analyses.slursAnalyzed = True

        output: bool = True
        output = output and self.analyzeSlursOrPhrases(HumdrumToken.SLUR, '**kern')
        output = output and self.analyzeSlursOrPhrases(HumdrumToken.SLUR, '**mens')
        return output

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzePhrasings -- Link start and ends of
    //    phrases to each other.
    '''
    def analyzePhrasings(self) -> bool:
        if self._analyses.phrasesAnalyzed:
            return False

        self._analyses.phrasesAnalyzed = True

        output: bool = True
        output = output and self.analyzeSlursOrPhrases(HumdrumToken.PHRASE, '**kern')
        return output

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeMensSlurs -- Link start and ends of
    //    slurs to each other.  They are the same as **kern, so borrowing
    //    analyzeKernSlurs to do the analysis.

        analyzeKernSlurs, analyzeMensSlurs, and analyzeKernPhrases (and their
        support functions) are almost completely identical, so I am combining
        them all into one parameterized set of functions. --gregc 17feb2021

        analyzeSlursOrPhrases takes two params:
            slurOrPhrase - HumdrumToken.SLUR or HumdrumToken.PHRASE
            spineDataType - '**kern' or '**mens'

            SLUR and PHRASE are defined (and used) in HumdrumToken.py, where I have
            combined the {slur,phrase}{Start,End}ElisionLevel functions into the
            startElisionLevel(slurOrPhrase, index) and
            endElisionLevel(slurOrPhrase, index)
    '''
    def analyzeSlursOrPhrases(self, slurOrPhrase: str, spineDataType: str) -> bool:
        spineStarts: t.List[HumdrumToken] = self.spineStartListOfType(spineDataType)
        if not spineStarts:
            return True

        # labels: first is previous label, last is next label
        labels: t.List[TokenPair] = [TokenPair(None, None)] * self.lineCount
        l: t.List[t.Optional[HumdrumToken]] = [None] * self.lineCount

        for i, line in enumerate(self._lines):
            if not line.isInterpretation:
                continue
            token = line[0]
            if token is not None and token.text.startswith('*>') and '[' not in token.text:
                l[i] = token

        current: t.Optional[HumdrumToken] = None
        for i, label in enumerate(l):
            if label is not None:
                current = label
            labels[i].first = current

        current = None
        for i, label in enumerate(reversed(l)):
            if label is not None:
                current = label
            labels[i].last = current

        endings: t.List[int] = [0] * self.lineCount
        ending: int = 0
        for i, label in enumerate(l):
            if label is not None:
                ending = 0
                lastChar = label.text[-1]
                if lastChar.isdigit():
                    ending = int(lastChar)
            endings[i] = ending

        output: bool = True
        slurStarts: t.List[HumdrumToken] = []
        slurEnds: t.List[HumdrumToken] = []
        linkSignifier: str = self._signifiers.linked
        for spineStart in spineStarts:
            output = (
                output and self.analyzeSpineSlursOrPhrases(
                    slurOrPhrase,
                    spineStart,
                    slurStarts,
                    slurEnds,
                    labels,
                    endings,
                    linkSignifier)
            )

        self.createLinkedSlursOrPhrases(slurOrPhrase, slurStarts, slurEnds)
        return output

    def analyzeSpineSlursOrPhrases(
            self,
            slurOrPhrase: str,
            spineStart: HumdrumToken,
            linkStarts: t.List[HumdrumToken],
            linkEnds: t.List[HumdrumToken],
            labels: t.List[TokenPair],
            endings: t.List[int],
            linkSig: str
    ) -> bool:
        beginStr: str = ''
        endStr: str = ''
        endingBackTag: str = ''
        sideTag: str = ''
        durationTag: str = ''
        hangingTag: str = ''
        openIndexTag: str = ''

        if slurOrPhrase == HumdrumToken.SLUR:
            beginStr = '('
            endStr = ')'
            endingBackTag = 'endingSlurBack'
            sideTag = 'slurSide'
            durationTag = 'slurDuration'
            hangingTag = 'hangingSlur'
            openIndexTag = 'slurOpenIndex'
        elif slurOrPhrase == HumdrumToken.PHRASE:
            beginStr = '{'
            endStr = '}'
            endingBackTag = 'endingPhraseBack'
            sideTag = 'phraseSide'
            durationTag = 'phraseDuration'
            hangingTag = 'hangingPhrase'
            openIndexTag = 'phraseOpenIndex'
        else:
            return False

        # linked phrases/slurs handled separately, so generate an ignore sequence:
        ignoreBegin: str = linkSig + beginStr
        ignoreEnd: str = linkSig + endStr

        # tracktokens == the 2-D data list for the track,
        # arranged in layers with the second dimension.
        trackTokens: t.List[t.List[HumdrumToken]] = self.getTrackSequence(
            startToken=spineStart, options=self.OPT_DATA | self.OPT_NOEMPTY
        )

        # opens == list of phrase/slur openings for each track and elision level
        # first dimension: elision level (max: 4)
        # second dimension: layer number (max: 8)
        opens: t.List[t.List[t.List[HumdrumToken]]] = []
        for _el in range(0, 4):
            opens.append([])
            for _tr in range(0, 8):
                opens[-1].append([])

        openCount: int = 0
        closeCount: int = 0
        elision: int = 0

        for rowOfTokens in trackTokens:
            for track, token in enumerate(rowOfTokens):
                if not token.isData:
                    continue
                if token.isNull:
                    continue
                openCount = token.text.count('(')
                closeCount = token.text.count(')')

                for i in range(0, closeCount):
                    if self._isLinkedSlurOrPhraseEnd(slurOrPhrase, token, i, ignoreEnd):
                        linkEnds.append(token)
                        continue

                    elision = token.endElisionLevel(slurOrPhrase, i)
                    if elision < 0:
                        continue

                    if opens[elision][track]:
                        # list of slur/phrase opens is not empty
                        self._linkSlurOrPhraseEndpoints(
                            slurOrPhrase, opens[elision][track][-1], token
                        )
                        # remove that last slur/phrase opening from buffer
                        opens[elision][track] = opens[elision][track][:-1]
                    else:
                        # No starting slur/phrase marker to match to this slur/phrase end in the
                        # given track.
                        # Search for an open slur/phrase in another track:
                        found: bool = False
                        for iTrack in range(0, len(opens[elision])):
                            if opens[elision][iTrack]:
                                # list of slur/phrase opens is not empty
                                self._linkSlurOrPhraseEndpoints(slurOrPhrase,
                                                               opens[elision][iTrack][-1],
                                                               token)
                                # remove slur/phrase opening from buffer
                                opens[elision][iTrack] = opens[elision][iTrack][:-1]
                                found = True
                                break
                        if not found:
                            lineIndex: int = token.lineIndex
                            endNum: int = endings[lineIndex]
                            pIndex: int = -1
                            first: t.Optional[HumdrumToken] = labels[lineIndex].first
                            if first is not None:
                                pIndex = first.lineIndex - 1

                            endNumPre: int = -1
                            if pIndex >= 0:
                                endNumPre = endings[pIndex]

                            if endNumPre > 0 and endNum > 0 and endNumPre != endNum:
                                # This is a slur/phrase in an ending that starts at the
                                # start of an ending.
                                duration: HumNum = token.durationFromStart
                                first = labels[token.lineIndex].first
                                if first is not None:
                                    duration = opFrac(duration - first.durationFromStart)
                                token.setValue('auto', endingBackTag, 'true')
                                token.setValue('auto', sideTag, 'stop')
                                token.setValue('auto', durationTag, token.durationToEnd)
                            else:
                                # This is a slur/phrase closing that does not have a
                                # matching opening.
                                token.setValue('auto', hangingTag, 'true')
                                token.setValue('auto', sideTag, 'stop')
                                token.setValue('auto', openIndexTag, str(i))
                                token.setValue('auto', durationTag, token.durationToEnd)

                for i in range(0, openCount):
                    if self._isLinkedSlurOrPhraseBegin(slurOrPhrase, token, i, ignoreBegin):
                        linkStarts.append(token)
                        continue
                    elision = token.startElisionLevel(slurOrPhrase, i)
                    if elision < 0:
                        continue
                    opens[elision][track].append(token)

        # Mark unclosed slur/phrase starts:
        for elisionOpens in opens:
            for trackOpens in elisionOpens:
                for openToken in trackOpens:
                    openToken.setValue('', 'auto', hangingTag, 'true')
                    openToken.setValue('', 'auto', sideTag, 'start')
                    openToken.setValue('', 'auto', durationTag, openToken.durationFromStart)

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::createLinkedSlurs --  Currently assume that
    //    start and ends are matched.
    '''
    def createLinkedSlursOrPhrases(self,
            slurOrPhrase: str,
            linkStarts: t.List[HumdrumToken],
            linkEnds: t.List[HumdrumToken]) -> None:
        shortest: int = len(linkStarts)
        if shortest > len(linkEnds):
            shortest = len(linkEnds)
        if shortest == 0:
            # nothing to do
            return

        for i in range(0, shortest):
            self._linkSlurOrPhraseEndpoints(slurOrPhrase, linkStarts[i], linkEnds[i])

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::isLinkedSlurEnd --
    '''
    @staticmethod
    def _isLinkedSlurOrPhraseEnd(
            slurOrPhrase: str,
            token: HumdrumToken,
            index: int,
            pattern: str
    ) -> bool:
        endStr: str = ''
        if slurOrPhrase == HumdrumToken.SLUR:
            endStr = ')'
        elif slurOrPhrase == HumdrumToken.PHRASE:
            endStr = '}'
        else:
            return False

        if pattern is None or len(pattern) <= 1:
            return False

        counter: int = -1
        for i in range(0, len(token.text)):
            if token.text[i] == endStr:
                counter += 1

            if i == 0:
                # Can't have linked slur at starting index in string.
                continue

            if counter != index:
                continue

            startIndex: int = 1 - len(pattern) + 1
            if token.text[startIndex:].startswith(pattern):
                return True
            return False

        return False

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::isLinkedSlurBegin --
    '''
    @staticmethod
    def _isLinkedSlurOrPhraseBegin(
            slurOrPhrase: str,
            token: HumdrumToken,
            index: int,
            pattern: str
    ) -> bool:
        startStr: str = slurOrPhrase

        if pattern is None or len(pattern) <= 1:
            return False

        counter: int = -1
        for i in range(0, len(token.text)):
            if token.text[i] == startStr:
                counter += 1

            if i == 0:
                continue

            if counter != index:
                continue

            startIndex = i - len(pattern) + 1
            if pattern in token.text[startIndex:]:
                return True
            return False
        return False

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::linkSlurEndpoints --  Allow up to two slur starts/ends
    //      on a note.
    '''
    @staticmethod
    def _linkSlurOrPhraseEndpoints(
            slurOrPhrase: str,
            startTok: HumdrumToken,
            endTok: HumdrumToken
    ) -> None:
        if slurOrPhrase == HumdrumToken.SLUR:
            prefix = 'slur'
        else:
            prefix = 'phrase'

        durTag: str = prefix + 'Duration'
        endTag: str = prefix + 'EndId'
        startTag: str = prefix + 'StartId'
        startNumberTag: str = prefix + 'StartNumber'
        endNumberTag: str = prefix + 'EndNumber'
        startCountTag: str = prefix + 'StartCount'
        endCountTag: str = prefix + 'EndCount'



        startCount: int = startTok.getValueInt('auto', startCountTag) + 1
        openCount: int = startTok.text.count(slurOrPhrase)
        openEnumeration = openCount - startCount + 1

        if openEnumeration > 1:
            openSuffix: str = str(openEnumeration)
            endTag += openSuffix
            durTag += openSuffix

        endCount: int = endTok.getValueInt('auto', endCountTag) + 1
        closeEnumeration: int = endCount
        if closeEnumeration > 1:
            closeSuffix: str = str(closeEnumeration)
            startTag += closeSuffix
            startNumberTag += closeSuffix

        duration: HumNum = opFrac(endTok.durationFromStart - startTok.durationFromStart)

        startTok.setValue('auto', endTag, endTok)
        startTok.setValue('auto', 'id', startTok)
        startTok.setValue('auto', endNumberTag, closeEnumeration)
        startTok.setValue('auto', durTag, duration)
        startTok.setValue('auto', startCountTag, startCount)

        endTok.setValue('auto', startTag, startTok)
        endTok.setValue('auto', 'id', endTok)
        endTok.setValue('auto', startNumberTag, openEnumeration)
        endTok.setValue('auto', endCountTag, endCount)

#         print('SLUR: starts at {} (lineNum:{}, fieldNum:{}), ends at {} (lineNum:{},
#               fieldNum:{})'.format(startTok.text, startTok.lineNumber, startTok.fieldNumber,
#               endTok.text, endTok.lineNumber, endTok.fieldNumber), file=sys.stderr)

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeKernTies -- Link start and ends of
    //    ties to each other.
    '''
    def analyzeKernTies(self) -> bool:
        output: bool = True
        linkedTieStarts: t.List[t.Tuple[HumdrumToken, int]] = []
        linkedTieEnds: t.List[t.Tuple[HumdrumToken, int]] = []
        linkSignifier: str = self._signifiers.linked

        output = self.generateLinkedTieStartsAndEnds(linkedTieStarts, linkedTieEnds, linkSignifier)
        self.createLinkedTies(linkedTieStarts, linkedTieEnds)
        return output

    '''
    // Could be generalized to allow multiple grand-staff pairs by limiting
    // the search spines for linking (probably with *part indications).
    // Currently all spines are checked for linked ties.
    '''

    def generateLinkedTieStartsAndEnds(
            self,
            linkedTieStarts: t.List[t.Tuple[HumdrumToken, int]],
            linkedTieEnds: t.List[t.Tuple[HumdrumToken, int]],
            linkSignifier: str
    ) -> bool:
        # Use this in the future to limit to grand-staff search (or 3 staves for organ):
        # vector<vector<HTp> > tracktokens;
        # this->getTrackSeq(tracktokens, spinestart, OPT_DATA | OPT_NOEMPTY);

        # Only analyzing linked ties for now (others ties are handled without analysis in
        # HumdrumFile.createMusic21Stream, for example).
        if linkSignifier is None or linkSignifier == '':
            return True

        lstart: str = linkSignifier + '['
        lmiddle: str = linkSignifier + '_'
        lend: str = linkSignifier + ']'

        startDatabase: t.List[t.Tuple[t.Optional[HumdrumToken], int]] = [(None, -1)] * 400

        for line in self._lines:
            if not line.isData:
                continue

            for tok in line.tokens():
                if not tok.isKern:
                    continue
                if not tok.isData:
                    continue
                if tok.isNull:
                    continue
                if tok.isRest:
                    continue

                for subtokenIdx, subtokenStr in enumerate(tok.subtokens):
                    if lstart in subtokenStr:
                        startb40: int = Convert.kernToBase40(subtokenStr)
                        startDatabase[startb40] = (tok, subtokenIdx)

                    if lend in subtokenStr:
                        endb40: int = Convert.kernToBase40(subtokenStr)
                        endEntry: t.Tuple[t.Optional[HumdrumToken], int] = startDatabase[endb40]
                        if endEntry[0] is not None:
                            if t.TYPE_CHECKING:
                                assert isinstance(endEntry[0], HumdrumToken)
                            linkedTieStarts.append((endEntry[0], endEntry[1]))
                            linkedTieEnds.append((tok, subtokenIdx))
                            startDatabase[endb40] = (None, -1)

                    if lmiddle in subtokenStr:
                        middleb40: int = Convert.kernToBase40(subtokenStr)
                        middleEntry: t.Tuple[t.Optional[HumdrumToken], int] = (
                            startDatabase[middleb40]
                        )
                        if middleEntry[0] is not None:
                            if t.TYPE_CHECKING:
                                assert isinstance(middleEntry[0], HumdrumToken)
                            linkedTieStarts.append((middleEntry[0], middleEntry[1]))
                            linkedTieEnds.append((tok, subtokenIdx))
                        startDatabase[middleb40] = (tok, subtokenIdx)

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::createLinkedTies --
    '''
    def createLinkedTies(
            self,
            linkStarts: t.List[t.Tuple[HumdrumToken, int]],
            linkEnds: t.List[t.Tuple[HumdrumToken, int]]
    ) -> None:
        shortest = len(linkStarts)
        if shortest > len(linkEnds):
            shortest = len(linkEnds)
        if shortest == 0:
            # nothing to do
            return

        for i in range(0, shortest):
            self._linkTieEndpoints(
                linkStarts[i][0], linkStarts[i][1], linkEnds[i][0], linkEnds[i][1]
            )

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::linkTieEndpoints --
    '''
    @staticmethod
    def _linkTieEndpoints(
            tieStart: HumdrumToken,
            startSubtokenIdx: int,
            tieEnd: HumdrumToken,
            endSubtokenIdx: int
    ) -> None:
        durTag: str = 'tieDuration'
        startTag: str = 'tieStart'
        endTag: str = 'tieEnd'
        startNumTag: str = 'tieStartSubtokenNumber'
        endNumTag: str = 'tieEndSubtokenNumber'

        startSubtokenNumber: int = startSubtokenIdx + 1
        endSubtokenNumber: int = endSubtokenIdx + 1

        if tieStart.isChord:
            if startSubtokenNumber > 0:
                startSuffix: str = str(startSubtokenNumber)
                durTag += startSuffix
                endNumTag += startSuffix
                endTag += startSuffix

        if tieEnd.isChord:
            if endSubtokenNumber > 0:
                endSuffix: str = str(endSubtokenNumber)
                startTag += endSuffix
                startNumTag += endSuffix

        tieStart.setValue('auto', endTag, tieEnd)
        tieStart.setValue('auto', 'id', tieStart)
        if endSubtokenNumber > 0:
            tieStart.setValue('auto', endNumTag, str(endSubtokenNumber))

        tieEnd.setValue('auto', startTag, tieStart)
        tieEnd.setValue('auto', 'id', tieEnd)
        if startSubtokenNumber > 0:
            tieEnd.setValue('auto', startNumTag, str(startSubtokenNumber))

        duration: HumNum = opFrac(tieEnd.durationFromStart - tieStart.durationFromStart)
        tieStart.setValue('auto', durTag, duration)

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeRestPositions -- Calculate the vertical position
    //    of rests on staves with two layers.
    '''
    def analyzeRestPositions(self) -> None:
        # I will not be assigning implicit vertical rest positions to workaround
        # verovio's subsequent adjustments.  I will just be noting explicit vertical
        # rest positions.  --gregc
        self.checkForExplicitVerticalRestPositions()

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::checkForExplicitVerticalRestPositions -- Starting at the
    //     current layer, check all rests in the same track for vertical positioning.
    '''
    def checkForExplicitVerticalRestPositions(self) -> None:
        # default to treble clef
        defaultBaseline: int = Convert.kernClefToBaseline('*clefG2')
        baselines: t.List[int] = [defaultBaseline] * (self.maxTrack + 1)

        for line in self._lines:
            if line.isInterpretation:
                for tok in line.tokens():
                    if not tok.isKern:
                        continue
                    if not tok.isClef:
                        continue

                    baselines[tok.track] = Convert.kernClefToBaseline(tok.text)

            if not line.isData:
                continue

            for tok in line.tokens():
                if not tok.isKern:
                    continue
                if not tok.isRest:
                    continue

                self._checkRestForVerticalPositioning(tok, baselines[tok.track])

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::checkRestForVerticalPositioning -- Read any pitch information attached
    //     to a rest and convert to ploc/oloc values.

        Instead of ploc and oloc for MEI (which humlib does), compute and save what the
        stepShift for a music21 rest should be.
    '''
    @staticmethod
    def _checkRestForVerticalPositioning(rest: HumdrumToken, baseline: int) -> bool:
        m = re.search(r'([A-Ga-g]+)', rest.text)
        if m is None:
            return False

        pitch: str = m.group(1)
        b7: int = Convert.kernToBase7(pitch)
        if b7 < 0:
            # that wasn't really a pitch, no vertical positioning for you
            return False

        diff: int = (b7 - baseline) + 100
        if diff % 2 != 0:
            # force to every other diatonic step (staff lines)
            if rest.duration > 1:
                b7 -= 1
            else:
                b7 += 1

        # Instead of ploc and oloc for MEI (which humlib does), compute and save what the
        # stepShift for a music21 rest should be.
        # stepShift = 0 means "on the middle line", -1 means "in the first space below the
        # midline", +2 means "on the line above the midline", etc.
        # baseline is the base7 pitch of the lowest line in the staff -> stepShift = -4
        # b7 is the base7 pitch (including octave) of where the rest should go, so:

        # midline is 2 spaces + 2 lines (4 diatonic steps) above baseline
        midline: int = baseline + 4
        stepShift: int = b7 - midline
        rest.setValue('auto', 'stepShift', str(stepShift))

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeKernAccidentals -- Identify accidentals that
    //    should be printed (only in **kern spines) as well as cautionary
    //    accidentals (accidentals which are forced to be displayed but otherwise
    //    would not be printed.  Algorithm assumes that all secondary tied notes
    //    will not display their accidental across a system break.  Consideration
    //    about grace-note accidental display still needs to be done.
    '''
    def analyzeKernAccidentals(self) -> bool:
        # ottava marks must be analyzed first:
        self.analyzeOttavas()

        # We will mark a "visible accidental" in four (Humdrum) situations:
        # 1. there is an 'n' specified with no "hidden" mark (a 'y' immediately
        #    following the 'n', or a 'yy' anywhere in the token)
        # 2. there is a single 'X' immediately following the '#', '-', or 'n'
        # 3. There is an editorial accidental signifier character immediately following
        #    the '#', '-', or 'n'.  We will also obey the edittype in the signifier.
        # 4. There is a linked layout param (('N', 'acc') and ('A', 'vis')) to specify
        #    display of a specific accidental (or combination, such as 'n#')

        # These are a part of what Verovio does.  I believe the rest of what Verovio does
        # is clearly about rendering decisions based on key signature and previous accidentals
        # in the measure, and thus belong in a renderer, not in a converter.

        for line in self._lines:
            if not line.hasSpines:
                continue

            if not line.isData:
                continue

            for token in line.tokens():
                if not token.isKern:
                    continue

                if token.isNull:
                    continue

                if token.isRest:
                    continue

                for k, subtok in enumerate(token.subtokens):
                    accid: int = Convert.kernToAccidentalCount(subtok)
                    isHidden: bool = False
                    if 'yy' not in subtok:  # if note is hidden accidental hiding isn't necessary
                        if ('ny' in subtok or '#y' in subtok or '-y' in subtok):
                            isHidden = True

                    if accid == 0 and 'n' in subtok and not isHidden:
                        # if humdrum data specifies 'n', we'll put in a cautionary accidental
                        token.setValue('auto', str(k), 'cautionaryAccidental', 'true')
                        token.setValue('auto', str(k), 'visualAccidental', 'true')
                    elif 'XX' not in subtok:
                        # The accidental is not necessary. See if there is a single "X"
                        # immediately after the accidental which means to force it to
                        # display.
                        if '#X' in subtok or '-X' in subtok or 'nX' in subtok:
                            token.setValue('auto', str(k), 'cautionaryAccidental', 'true')
                            token.setValue('auto', str(k), 'visualAccidental', 'true')

                    # editorialAccidental analysis
                    isEditorial: bool = False
                    editType: str = ''
                    for x, signifier in enumerate(self._signifiers.editorialAccidentals):
                        if signifier in subtok:
                            isEditorial = True
                            editType = self._signifiers.editorialAccidentalTypes[x]
                            break

                    subTokenIdx: int = k
                    if len(token.subtokens) == 1:
                        subTokenIdx = -1
                    editType2: str = token.layoutParameter('A', 'edit', subTokenIdx)
                    if editType2 and not editType:
                        isEditorial = True
                        if editType2 == 'true':
                            # default editorial accidental type
                            editType = ''
                            # use the first editorial accidental RDF style in file if present
                            if self._signifiers.editorialAccidentalTypes:
                                editType = self._signifiers.editorialAccidentalTypes[0]
                        else:
                            editType = editType2

                    if isEditorial:
                        token.setValue('auto', str(k), 'editorialAccidental', 'true')
                        token.setValue('auto', str(k), 'visualAccidental', 'true')
                        if editType:
                            token.setValue('auto', str(k), 'editorialAccidentalStyle', editType)

                    # layout parameters (('N', 'acc') and ('A', 'vis')) can also be used
                    # to specify a specific accidental or combination to be made visible.
                    # 'A', 'vis' takes priority over 'N', 'acc', if both are present
                    layoutAccidental: str = token.layoutParameter('A', 'vis', subTokenIdx)
                    if not layoutAccidental:
                        layoutAccidental = token.layoutParameter('N', 'acc', subTokenIdx)
                    if layoutAccidental:
                        token.setValue('auto', str(k), 'cautionaryAccidental', layoutAccidental)
                        token.setValue('auto', str(k), 'visualAccidental', layoutAccidental)

                    # check for accidentals on trills, mordents, and turns
#                     if 't' in subtok:
#                     elif 'T' in subtok:
#                     elif 'M' in subtok:
#                     elif 'm' in subtok:
#                     elif 'W' in subtok:
#                     elif 'w' in subtok:
#                     elif '$' in subtok:
#                     elif 'S' in subtok:

        # Indicate that the accidental analysis has been done:
        self.setValue('auto', 'accidentalAnalysis', 'true')

        return True

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeOttavas --

        analyzeOttavas reads any 8va-type data, and makes a note of it in each affected token.
        Q: octaveState assumes no nesting of ottavas, but activeOttava is all about
        Q: handling nesting of ottavas. --gregc
    '''
    def analyzeOttavas(self) -> None:
        trackCount: int = self.maxTrack
        activeOttava: t.List[int] = [0] * (trackCount + 1)  # 0th element goes unused
        octaveState: t.List[int] = [0] * (trackCount + 1)   # 0th element goes unused

        ttrack: int

        for line in self._lines:
            if line.isInterpretation:
                for token in line.tokens():
                    if not token.isKern:
                        continue
                    ttrack = token.track
                    if token.text == '*8va':
                        octaveState[ttrack] = +1
                        activeOttava[ttrack] += 1
                    elif token.text == '*X8va':
                        octaveState[ttrack] = 0
                        activeOttava[ttrack] -= 1
                    elif token.text == '*8ba':
                        octaveState[ttrack] = -1
                        activeOttava[ttrack] += 1
                    elif token.text == '*X8ba':
                        octaveState[ttrack] = 0
                        activeOttava[ttrack] -= 1
                    elif token.text == '*15ma':
                        octaveState[ttrack] = +2
                        activeOttava[ttrack] += 1
                    elif token.text == '*X15ma':
                        octaveState[ttrack] = 0
                        activeOttava[ttrack] -= 1
                    elif token.text == '*15ba':
                        octaveState[ttrack] = -2
                        activeOttava[ttrack] += 1
                    elif token.text == '*X15ba':
                        octaveState[ttrack] = 0
                        activeOttava[ttrack] -= 1
            elif line.isData:
                for token in line.tokens():
                    if not token.isKern:
                        continue
                    ttrack = token.track
                    if activeOttava[ttrack] == 0:  # if nesting level is 0
                        continue
                    if octaveState[ttrack] == 0:  # if octave adjustment is 0
                        continue
                    if token.isNull:
                        continue
                    # do not exclude rests, since the vertical placement of the rest
                    # on the staff may need to be updated by the ottava mark.
                    # if token.isRest:
                    #     continue

                    token.setValue('auto', 'ottava', str(octaveState[ttrack]))

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeTextRepetition -- Look for *ij and *Xij markers
    //     that indicate repetition marks.  values added to text:
    //      	auto/ij=true: the syllable is in an ij region.
    //      	auto/ij-begin=true: the syllable is the first in an ij region.
    //      	auto/ij-end=true: the syllable is the last in an ij region.
    //
    // Returns true if there are any *ij/*Xij markers in the data.
    '''
    def analyzeTextRepetition(self) -> None:
        spineStarts: t.List[t.Optional[HumdrumToken]] = self.spineStartList

        for start in spineStarts:
            ijstate: bool = False
            startij: bool = False
            lastIJToken: t.Optional[HumdrumToken] = None  # last token seen in ij range

            if start is None:
                raise HumdrumInternalError('start that is None found in spineStarts')

            # BUGFIX: **sylb -> **silbe
            if not start.isDataType('**text') and not start.isDataType('**silbe'):
                continue

            current: t.Optional[HumdrumToken] = start
            while current is not None:
                if current.isNull:
                    current = current.nextToken0
                    continue

                if current.isInterpretation:
                    if current.text == '*ij':
                        startij = True
                        ijstate = True
                    elif current.text == '*Xij':
                        startij = False
                        ijstate = False
                        if lastIJToken is not None:
                            lastIJToken.setValue('auto', 'ij-end', 'true')
                            lastIJToken = None
                    current = current.nextToken0
                    continue

                if current.isData:
                    if ijstate:
                        current.setValue('auto', 'ij', 'true')
                        if startij:
                            current.setValue('auto', 'ij-begin', 'true')
                            startij = False
                        lastIJToken = current

                current = current.nextToken0

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::analyzeRScale --
    '''
    def analyzeRScale(self) -> bool:
        numActiveTracks: int = 0  # number of tracks currently having an active rscale parameter
        rscales: t.List[HumNum] = [opFrac(1)] * (self.maxTrack + 1)
        ttrack: int

        for line in self._lines:
            if line.isInterpretation:
                for token in line.tokens():
                    if not token.isKern:
                        continue

                    if not token.text.startswith('*rscale:'):
                        continue

                    value: HumNum = opFrac(1)
                    m = re.search(r'\*rscale:(\d+)/(\d+)', token.text)
                    if m is not None:
                        top: int = int(m.group(1))
                        bot: int = int(m.group(2))
                        value = opFrac(Fraction(top, bot))
                    else:
                        m = re.search(r'\*rscale:(\d+)', token.text)
                        if m is not None:
                            top = int(m.group(1))
                            value = opFrac(top)

                    ttrack = token.track
                    if value == 1:
                        if rscales[ttrack] != 1:
                            rscales[ttrack] = opFrac(1)
                            numActiveTracks -= 1
                    else:
                        if rscales[ttrack] == 1:
                            numActiveTracks += 1
                        rscales[ttrack] = value
                continue

            if numActiveTracks == 0:
                continue

            if not line.isData:
                continue

            for token in line.tokens():
                ttrack = token.track
                if rscales[ttrack] == 1:
                    continue
                if not token.isKern:
                    continue
                if token.isNull:
                    continue
                if token.duration < 0:
                    continue

                dur: HumNum = opFrac(token.durationNoDots * rscales[ttrack])
                vis: str = Convert.durationToRecip(dur)
                vis += '.' * token.dotCount
                token.setValue('LO', 'N', 'vis', vis)

        return True

    '''
    //////////////////////////////
    //
    // HumdrumInput::initializeSpineColor -- Look for *color: interpretations before data.

        This only looks at the beginning of the file, before any data tokens.
        Later, we will update self._spineColor when necessary as we process the file.

    '''
    def initializeSpineColor(self) -> None:
        self._spineColor = [[''] * MAXCOLORSUBTRACK] * (self.maxTrack + 1)
        for line in self.lines():
            if line.isData:
                break

            if not line.isInterpretation:
                continue

            if '*color:' not in line.text:
                continue

            for token in line.tokens():
                self.setSpineColorFromColorInterpToken(token)

    '''
        setSpineColorFromColorInterpToken - factored from several places in HumdrumFileContent.py
        and HumdrumFile.py
    '''
    def setSpineColorFromColorInterpToken(self, token: HumdrumToken) -> None:
        m = re.search(r'^\*color:(.*)', token.text)
        if m:
            ctrack: t.Optional[int] = token.track
            strack: t.Optional[int] = token.subTrack
            if ctrack is None or strack is None:
                return
            if strack < MAXCOLORSUBTRACK:
                self._spineColor[ctrack][strack] = m.group(1)
                if strack == 1:
                    # copy it to subtrack 0 as well
                    self._spineColor[ctrack][0] = m.group(1)
                elif strack == 0:
                    # copy it to all subtracks
                    for z in range(1, MAXCOLORSUBTRACK):
                        self._spineColor[ctrack][z] = m.group(1)

    '''
    //////////////////////////////
    //
    // HumdrumFileContent::hasPickup -- Return false if there is no pickup measure.
    //   Return the barline index number if there is a pickup measure.  A pickup measure
    //   is identified when the duration from the start of the file to the first
    //   barline is not zero or equal to the duration of the starting time signature.
    //   if there is not starting time signature, then there cannot be an identified
    //   pickup measure.
    '''
    def hasPickup(self) -> int:
        barlineIdx: int = -1
        tsig: t.Optional[HumdrumToken] = None

        for i, line in enumerate(self._lines):
            if line.isBarline:
                if barlineIdx > 0:
                    # second barline found, so stop looking for time signature
                    break
                barlineIdx = i
                continue

            if not line.isInterpretation:
                continue

            if tsig is not None:
                continue

            for token in line.tokens():
                if token.isTimeSignature:
                    tsig = token
                    break

        if tsig is None:
            # no time signature so return 0
            return 0

        if barlineIdx < 0:
            # no barlines in music
            return 0

        mdur: HumNum = self._lines[barlineIdx].durationFromStart
        tdur: HumNum = Convert.timeSigToDuration(tsig)
        if mdur == tdur:
            # first measure is exactly as long as the time signature says: no pickup measure
            return 0

        return barlineIdx
